// Copyright 2016 Orus.io. All rights reserved.

package auth

import (
	"encoding/pem"
	"fmt"
	"regexp"
	"runtime/debug"
	"strconv"
	"strings"
	"time"

	"github.com/nats-io/gnatsd/server"
	nats "github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"
	"github.com/rs/zerolog"

	"xbus.io/go-xbus/v3/api"
	"xbus.io/xbus/storage"
)

var apiVersion ProtocolVersion

// ProtocolVersion is a parsed protocol version
type ProtocolVersion struct {
	Major int
	Minor int
}

func (v ProtocolVersion) String() string {
	return strconv.Itoa(v.Major) + "." + strconv.Itoa(v.Minor)
}

// ParseProtocolVersion parses a protocol version
func ParseProtocolVersion(version string) (*ProtocolVersion, error) {
	splitted := strings.Split(version, ".")
	if len(splitted) != 2 {
		return nil, fmt.Errorf("Invalid protocol version: %s", version)
	}
	major, err := strconv.Atoi(splitted[0])
	if err != nil {
		return nil, fmt.Errorf(
			"Invalid protocol version: %s. '%s' is not a valid number",
			version, splitted[0])
	}
	minor, err := strconv.Atoi(splitted[1])
	if err != nil {
		return nil, fmt.Errorf(
			"Invalid protocol version: %s. '%s' is not a valid number",
			version, splitted[1])
	}
	return &ProtocolVersion{
		Major: major,
		Minor: minor,
	}, nil
}

// ClientInfo is a parsed client information (found in the 'Name' connect info field)
type ClientInfo struct {
	Name            string
	ProtocolVersion ProtocolVersion
	Client          string
	ClientVersion   string
}

var (
	clientInfoRe = regexp.MustCompile(`([^@]*)@([0-9]*\.[0-9]*)(\(([^@]*)(@([^\)]*))?\))?`)
)

// ParseClientInfo parses a 'Name' connect info field
func ParseClientInfo(name string) (*ClientInfo, error) {
	match := clientInfoRe.FindStringSubmatch(name)
	if len(match) == 0 {
		return nil, fmt.Errorf(
			"client info should match '%s' - got '%s'", clientInfoRe.String(), name)
	}
	version, err := ParseProtocolVersion(match[2])
	if err != nil {
		return nil, err
	}
	return &ClientInfo{
		Name:            match[1],
		ProtocolVersion: *version,
		Client:          match[4],
		ClientVersion:   match[6],
	}, nil
}

// NewXBusAuth returns a gnatsd.server.Auth implementation on
// top of actor & account storage instances.
// It provides a simple login/password authentication based on the Account model stored in the
// database.
// A special token allow full-access (used by internal services)
func NewXBusAuth(
	accountStorage storage.AccountStorage,
	passwordStorage storage.PasswordStorage,
	actorStorage storage.ActorStorage,
	sessionStorage storage.SessionStorage,
	logger zerolog.Logger,
) XBusAuth {
	return XBusAuth{
		AccountStorage:  accountStorage,
		PasswordStorage: passwordStorage,
		ActorStorage:    actorStorage,
		SessionStorage:  sessionStorage,
		AuthToken:       nuid.Next(),
		Logger:          logger,
	}
}

// XBusAuth ...
type XBusAuth struct {
	AccountStorage  storage.AccountStorage
	PasswordStorage storage.PasswordStorage
	ActorStorage    storage.ActorStorage
	SessionStorage  storage.SessionStorage
	AuthToken       string
	Logger          zerolog.Logger
}

func defaultPermissions(token string) *server.Permissions {
	return &server.Permissions{
		Publish: &server.SubjectPermission{
			Allow: []string{api.XbusAPIPrefix + "registration.register"},
		},
		Subscribe: &server.SubjectPermission{
			Allow: []string{
				nats.InboxPrefix + "anon-" + token + ".in.*",
			},
		},
	}
}

// Check authenticate a client for first or third auth step
func (a *XBusAuth) Check(c server.ClientAuthentication) bool {
	// We do not want storage errors to crash the server
	defer func() {
		if r := recover(); r != nil {
			a.Logger.Error().Interface("error", r).
				Str("stack", string(debug.Stack())).
				Msg("unexpected error in identity Check")
		}
	}()

	tlsState := c.GetTLSConnectionState()

	if tlsState == nil {
		// non-TLS connections are no longer allowed
		remote := c.RemoteAddress()
		a.Logger.Error().
			Str("remote-addr", remote.String()).
			Msg("Non-TLS connection attempt")
		return false
	}

	opts := c.GetOpts()

	clientInfo, err := ParseClientInfo(opts.Name)
	if err != nil {
		// TODO if we want pre-3.0.0 versions to connect, we may allow an empty "Name"
		a.Logger.Err(err).Msg("error parsing client info")
		return false
	}

	if clientInfo.ProtocolVersion.Major != apiVersion.Major {
		a.Logger.Error().
			Str("expected-protocol-version", strconv.Itoa(apiVersion.Major)+".x").
			Str("client-protocol-version", clientInfo.ProtocolVersion.String()).
			Msg("Incompatible client version")
		return false
	}

	if a.AuthToken != "" && opts.Authorization == a.AuthToken {
		return true
	}

	if len(tlsState.VerifiedChains) == 0 &&
		opts.Password == "" &&
		!strings.HasPrefix(opts.Authorization, "session#") {
		a.Logger.Info().Msg(
			"No client certificate & no password -> anonymous connection")
		user := server.User{
			Username:    "anonymous",
			Permissions: defaultPermissions(opts.Username),
		}
		c.RegisterUser(&user)
		return true
	}

	var account api.Account
	var ok bool

	if strings.HasPrefix(opts.Authorization, "session#") {
		session, err := a.SessionStorage.Get(opts.Authorization[8:])
		if err != nil {
			a.Logger.Err(err).Msg("error getting the session")
			return false
		}
		if session == nil {
			a.Logger.Error().
				Str("session-token", opts.Authorization[8:]).
				Msg("unknown session token")
			return false
		}
		fmt.Println("Valid until", session.ValidUntil, "- now", time.Now())
		if session.ValidUntil.Before(time.Now()) {
			a.Logger.Error().Time("valid-until", session.ValidUntil).
				Msg("session expired")
			return false
		}

		account, ok = a.AccountStorage.Get(session.AccountID)
		if !ok {
			a.Logger.Error().
				Str("session-token", session.Token).
				Str("account-id", session.AccountID.String()).
				Msg("session is associated to an unknown account")
			return false
		}
	} else if opts.Password != "" {
		a.Logger.Info().Str("username", opts.Username).Msg("will do a password auth")
		// find the account
		account, ok = a.AccountStorage.GetByName(opts.Username)
		if !ok {
			a.Logger.Error().Str("username", opts.Username).Msg("unknown account")
			return false
		}
		ok = a.PasswordStorage.Verify(account.ID, opts.Password)
		if !ok {
			a.Logger.Error().Str("username", opts.Username).Msg("invalid password")
		}
	} else {
		for _, chain := range tlsState.VerifiedChains {
			cert := chain[0]
			pCert := pem.EncodeToMemory(&pem.Block{
				Type:  "CERTIFICATE",
				Bytes: cert.Raw,
			})
			account, ok = a.AccountStorage.GetByCert(string(pCert))
			if ok {
				a.Logger.Debug().Str("account-name", account.Name).
					Msg("found account matching certificate")
				break
			}
		}
		if !ok {
			a.Logger.Error().Msg("unknown certificate")
		}
	}

	if !ok {
		return false
	}

	var apiKey string

	if account.Type == api.Account_GATEWAY {
		if strings.HasPrefix(opts.Authorization, "-----BEGIN CERTIFICATE-----") {
			account, ok = a.AccountStorage.GetByCert(opts.Authorization)
		} else if opts.Authorization == "gateway" {
			// keep the gateway account, it has a few permissions
		} else if strings.HasPrefix(opts.Authorization, "webhook:") {
			account, ok = a.AccountStorage.Get(api.MustUUIDFromString(opts.Authorization[8:]))
		} else if opts.Authorization != "" {
			apiKey = opts.Authorization
			account, ok = a.AccountStorage.GetByAPIKey(opts.Authorization)
		}
		if !ok {
			a.Logger.Error().Str("auth_token", opts.Authorization).Msg("auth_token matches no account")
			return false
		}
	}

	whoamiSubject, whoamiReply := api.WhoAmISubjects(account.Name)
	whoamiSubject2, whoamiReply2 := api.WhoAmISubjects(account.ID.String())

	// Default permissions any account has.
	permissions := server.Permissions{
		Publish: &server.SubjectPermission{
			Allow: []string{
				whoamiSubject,
				whoamiSubject + ".json",
				whoamiSubject2,
				whoamiSubject2 + ".json",
				nats.InboxPrefix + account.ID.String() + ".out.*",
				nats.InboxPrefix + account.ID.String() + ".in.*.heartbeat",
				api.ClientPrefix + account.ID.String() + ".>",
				api.ClientPrefix + account.Name + ".>",
				api.XbusAPIPrefix + "registration.update." + account.ID.String(),
			},
		},
		Subscribe: &server.SubjectPermission{
			Allow: []string{
				whoamiReply,
				whoamiReply2,
				nats.InboxPrefix + account.ID.String() + ".in.*",
				nats.InboxPrefix + account.ID.String() + ".out.*.heartbeat",
			},
		},
	}

	if apiKey != "" {
		whoamiSubject, whoamiReply := api.WhoAmISubjects(apiKey)
		permissions.Publish.Allow = append(
			permissions.Publish.Allow,
			whoamiSubject,
			whoamiSubject+".json")
		permissions.Subscribe.Allow = append(
			permissions.Subscribe.Allow,
			whoamiReply,
		)
	}

	switch account.Type {
	case api.Account_ACTOR:
		err = fillActorPermissions(account, &permissions, a)
	case api.Account_USER:
		err = fillUserPermissions(account, &permissions)
	case api.Account_GATEWAY:
		err = fillGatewayPermissions(account, &permissions)
	default:
		a.Logger.Error().Str("account-type", account.Type.String()).Msg("invalid account type")
		return false
	}

	if err != nil {
		a.Logger.Err(err).Msg("error filling permissions")
		return false
	}

	a.Logger.Debug().Str("account-name", account.Name).
		Interface("permissions", permissions).
		Msg("")
	user := server.User{Username: account.Name, Permissions: &permissions}
	c.RegisterUser(&user)
	a.Logger.Info().Msgf("authenticated %s", account.Name)

	return true
}

// Fill permissions Xbus actors have access to.
func fillActorPermissions(
	account api.Account, permissions *server.Permissions, a *XBusAuth) error {

	// Got an account; find corresponding actors to set permissions accordingly.
	actors := a.ActorStorage.ListAccount(account.ID)

	// Add specific permissions for each actor.
	for _, actor := range actors {
		switch actor.Kind {
		case api.Actor_EMITTER:
			permissions.Publish.Allow = append(
				permissions.Publish.Allow,
				api.ActorOutputSubject(actor.ID),
				api.ActorOutputSubject(actor.ID)+".json",
				api.XbusAPIPrefix+"processstate.getenvelopestate."+actor.ID.String(),
				api.XbusAPIPrefix+"processstate.getenvelopestate."+actor.ID.String()+".*",
				api.XbusAPIPrefix+"processstate.ackresult."+actor.ID.String(),
				api.XbusAPIPrefix+"processstate.ackresult."+actor.ID.String()+".*",
				api.XbusAPIPrefix+"processstate.list."+actor.ID.String(),
				api.XbusAPIPrefix+"processstate.list."+actor.ID.String()+".*",
			)
			permissions.Subscribe.Allow = append(
				permissions.Subscribe.Allow,
				api.EmitterEnvelopeStateSubject(actor.ID, api.UUID{}),
				api.EmitterEnvelopeStateSubject(actor.ID, api.UUID{})+".json",
			)
		case api.Actor_CONSUMER:
			permissions.Subscribe.Allow = append(
				permissions.Subscribe.Allow,
				api.XbusAPIPrefix+"actoragent."+actor.ID.String()+".>",
			)
			permissions.Publish.Allow = append(
				permissions.Publish.Allow,
				api.XbusAPIPrefix+"envelopestorage.retrieve.actor."+actor.ID.String(),
				api.XbusAPIPrefix+"envelopestorage.retrieve.actor."+actor.ID.String()+".*",
			)
		case api.Actor_WORKER:
			permissions.Subscribe.Allow = append(
				permissions.Subscribe.Allow,
				api.XbusAPIPrefix+"actoragent."+actor.ID.String()+".>",
			)
			permissions.Publish.Allow = append(
				permissions.Publish.Allow,
				api.ActorOutputSubject(actor.ID),
				api.XbusAPIPrefix+"envelopestorage.retrieve.actor."+actor.ID.String(),
				api.XbusAPIPrefix+"envelopestorage.retrieve.actor."+actor.ID.String()+".*",
			)
		default:
			return fmt.Errorf("Invalid actor type: %s", actor.Kind)
		}
		permissions.Publish.Allow = append(
			permissions.Publish.Allow,
			api.XbusAPIPrefix+"sentinel.actorready."+actor.ID.String(),
			api.XbusAPIPrefix+"sentinel.actorleaving."+actor.ID.String(),
			api.XbusAPIPrefix+"director.*."+actor.ID.String(),
			api.XbusAPIPrefix+"director.*."+actor.ID.String()+".*",
		)
		permissions.Subscribe.Allow = append(
			permissions.Subscribe.Allow,
			api.XbusAPIPrefix+"sentinel.survey",
		)
	}
	return nil
}

// Fill permissions regular API users have access to.
func fillUserPermissions(account api.Account, permissions *server.Permissions) error {
	permissions.Publish.Allow = append(
		permissions.Publish.Allow,
		api.XbusAPIPrefix+"control.>",
		api.XbusAPIPrefix+"envelopestorage.retrieve.account."+account.ID.String(),
		api.XbusAPIPrefix+"envelopestorage.retrieve.account."+account.ID.String()+".*",
	)
	return nil
}

func fillGatewayPermissions(account api.Account, permissions *server.Permissions) error {
	permissions.Publish.Allow = append(permissions.Publish.Allow, api.XbusAPIPrefix+"control.actor.query")
	return nil
}

func init() {
	v, err := ParseProtocolVersion(api.ProtocolVersion)
	if err != nil {
		panic(err)
	}
	apiVersion = *v
}
