package cmd

import (
	"fmt"
	"io"
	"os"

	"github.com/rs/zerolog"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v3"
)

// Logger is the default logger for all commands.
// it is initialized early by initConfig(), so any code called
// by the Run function of a command is safe to use it.
var Logger xbus.ZeroLogger

func autoColors() bool {
	stat, err := os.Stdout.Stat()
	if err != nil || (stat.Mode()&os.ModeCharDevice) == 0 {
		return false
	}
	return true
}

func initLogger(status bool, outputMiddleware func(io.Writer) io.Writer) {
	var (
		appName     = "xbusd"
		debug       = viper.GetBool("debug")
		trace       = viper.GetBool("trace")
		quiet       = viper.GetBool("quiet")
		logLevelArg = viper.GetString("log-level")
		logLevel    = zerolog.InfoLevel
		errMsg      string
	)
	switch {
	case trace:
		logLevel = zerolog.TraceLevel
	case debug:
		logLevel = zerolog.DebugLevel
	case quiet:
		logLevel = zerolog.ErrorLevel
	case logLevelArg == "trace":
		logLevel = zerolog.TraceLevel
	case logLevelArg == "debug":
		logLevel = zerolog.DebugLevel
	case logLevelArg == "info":
		logLevel = zerolog.InfoLevel
	case logLevelArg == "warn", logLevelArg == "warning":
		logLevel = zerolog.WarnLevel
	case logLevelArg == "error":
		logLevel = zerolog.ErrorLevel
	case logLevelArg == "fatal":
		logLevel = zerolog.FatalLevel
	default:
		errMsg = fmt.Sprintf("invalid log-level: %s", logLevelArg)
	}

	var out io.Writer = os.Stderr

	if autoColors() {
		out = zerolog.ConsoleWriter{Out: out}
	}

	if outputMiddleware != nil {
		out = outputMiddleware(out)
	}

	log := zerolog.
		New(out).
		With().
		Timestamp().
		Str("app", appName).
		Logger().
		Level(logLevel)

	Logger = xbus.WrapZeroLogger(log)
	if errMsg != "" {
		log.Fatal().Msg(errMsg)
	}
	if status {
		Logger.Debugf("Logger ready. loglevel=%v", logLevel)
	}
}
