package fullstack

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	"xbus.io/xbus/testutils"
)

const graph1 = `
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  actors: [consumer1]
  inputs: [default]
edges:
- source -> target
`

const graph2 = `
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  actors: [consumer2]
  inputs: [default]
edges:
- source -> target
`

func TestProcessReplay(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	consumer1Hit := make(chan *api.Envelope, 2)
	consumer2Hit := make(chan *api.Envelope, 2)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	_, consumer1 := ft.CreateConsumer("consumer1", []string{"anyconsumer"})
	xsub1, err := consumer1.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			e, err := apc.CompleteEnvelope("default", time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			consumer1Hit <- e
			return nil
		})
	assert.Nil(t, err)
	defer testutils.DeferedAssertNoError(t, xsub1.Unsubscribe)

	_, consumer2 := ft.CreateConsumer("consumer2", []string{"anyconsumer"})
	xsub2, err := consumer2.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			e, err := apc.CompleteEnvelope("default", time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			consumer2Hit <- e
			return nil
		})
	assert.Nil(t, err)
	defer testutils.DeferedAssertNoError(t, xsub2.Unsubscribe)

	ft.InstallNamedGraph("default", "1", graph1)

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	var processID api.UUID

	timeout := time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated")
	case envState := <-envStateChan:
		processID = envState.ProcessID
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-consumer1Hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated to 'done'")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_DONE, envState.Status)
	}

	ft.InstallNamedGraph("default", "2", graph2)

	// Now let's replay the process. Once with the same pipeline, once with
	// a new match
	p1, err := ft.Client.Control.Process.Replay(
		control.ProcessReplayRequest{
			Process: api.Process{ID: processID},
			Match:   false,
		},
	)
	if !assert.NoError(t, err) {
		t.FailNow()
	}
	assert.Equal(t, processID, p1.GroupID)
	assert.Equal(t, processID, p1.ReplayOf)

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-consumer1Hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	p2, err := ft.Client.Control.Process.Replay(
		control.ProcessReplayRequest{
			Process: api.Process{ID: p1.ID},
			Match:   true,
		},
	)
	if !assert.NoError(t, err) {
		t.FailNow()
	}
	assert.Equal(t, processID, p2.GroupID)
	assert.Equal(t, p1.ID, p2.ReplayOf)

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-consumer2Hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}
