// Copyright 2019 Orus.io. All rights reserved.

package metrics

import (
	"context"
	"crypto/x509"
	"errors"
	"fmt"
	"math"
	"sync"
	"time"

	"github.com/getsentry/sentry-go"
	"github.com/prometheus/client_golang/prometheus"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	xbus_tools "xbus.io/go-xbus/v3/tools"
	"xbus.io/xbus/services"
	"xbus.io/xbus/services/tools"
)

const (
	// MetricsQueue is the name we use to queueSub to a queue with the metrics
	MetricsQueue = "metrics_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options

	CertificateExpirationCheckInterval    time.Duration
	CertificateExpirationNotifyInterval   time.Duration
	CertificateExpirationWarningThreshold time.Duration
	CertificateExpirationErrorThreshold   time.Duration

	ServerCA []byte
	ClientCA []byte
	Hub      *sentry.Hub
}

// NewService returns a new Service
func (o Options) NewService() (*Service, error) {
	if o.CertificateExpirationCheckInterval == 0 {
		o.CertificateExpirationCheckInterval = time.Minute
	}
	if o.CertificateExpirationNotifyInterval == 0 {
		o.CertificateExpirationNotifyInterval = time.Hour
	}
	if o.CertificateExpirationWarningThreshold == 0 {
		o.CertificateExpirationWarningThreshold = time.Hour * 24 * 28
	}
	if o.CertificateExpirationErrorThreshold == 0 {
		o.CertificateExpirationErrorThreshold = time.Hour * 24 * 7
	}
	serverCA, err := x509.ParseCertificate(o.ServerCA)
	if err != nil {
		o.Logger.Err(err).Msg("Could not parse server ca")
		return nil, err
	}
	clientCA, err := x509.ParseCertificate(o.ClientCA)
	if err != nil {
		o.Logger.Err(err).Msg("Could not parse client ca")
		return nil, err
	}

	registry := prometheus.NewRegistry()

	return &Service{
		BaseService: services.NewBaseService("metrics", o.Options),
		Opts:        o,

		serverCAValidNotAfter: serverCA.NotAfter,
		clientCAValidNotAfter: clientCA.NotAfter,

		registry:   registry,
		registerer: registry,

		// metrics
		actorOnlineStatus: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Name: "xbus_server_actor_online_status",
			Help: "Consumer/worker online status. 0 if the actor is offline, 1 if the actor is online",
		}, []string{"name"}),

		accountDaysToCertificateExpiration: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Name: "xbus_server_account_days_to_certificate_expiration",
			Help: "Days until actor certificate expiration",
		}, []string{"name"}),

		serverDaysToCertificateExpiration: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Name: "xbus_server_days_to_certificate_expiration",
			Help: "Days until a server certificate expiration",
		}, []string{"name"}),

		processCompletionTime: prometheus.NewSummaryVec(prometheus.SummaryOpts{
			Name:       "xbus_process_completion_time_seconds",
			Help:       "Process completion time, in seconds",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		}, []string{"pipeline", "pipeline_version", "emitter", "status"}),
	}, nil
}

// New returns a new Service
func New(options Options) (*Service, error) {
	return options.NewService()
}

// Service is the metrics service implementation
// The metrics service is responsible for keeping track of global metrics
// of the bus
type Service struct {
	services.BaseService
	Opts Options

	ctx       context.Context
	stop      func()
	wg        sync.WaitGroup
	api       *api.Client
	control   *control.Client
	actorMap  *tools.ActorMap
	pipelines *tools.PipelineCache
	actors    *tools.ActorCache

	serverCAValidNotAfter time.Time
	clientCAValidNotAfter time.Time

	// metrics
	registry   *prometheus.Registry
	registerer prometheus.Registerer

	actorOnlineStatus                  *prometheus.GaugeVec
	accountDaysToCertificateExpiration *prometheus.GaugeVec
	serverDaysToCertificateExpiration  *prometheus.GaugeVec
	processCompletionTime              *prometheus.SummaryVec
	processCount                       *SummaryCollector
	envelopeStorageMetrics             *EnvelopeStatCollector
	logStorageMetrics                  *LogStatCollector
}

// Gatherer returns the prometheus registry
func (s *Service) Gatherer() prometheus.Gatherer {
	return s.registry
}

// Startup ...
func (s *Service) Startup() error {

	if err := s.registerer.Register(s.accountDaysToCertificateExpiration); err != nil {
		return err
	}
	if err := s.registerer.Register(s.actorOnlineStatus); err != nil {
		return err
	}
	if err := s.registerer.Register(s.serverDaysToCertificateExpiration); err != nil {
		return err
	}
	if err := s.registerer.Register(s.processCompletionTime); err != nil {
		return err
	}

	s.ctx, s.stop = context.WithCancel(context.Background())
	s.api = api.NewClient(s.Conn)
	s.control = control.NewClient(s.Conn)
	if m, err := tools.NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}
	s.pipelines = tools.NewPipelineCache(s.control)
	s.actors = tools.NewActorCache(s.control)

	s.processCount = NewSummaryCollector(s.control, s.actors, s.pipelines)
	if err := s.registerer.Register(s.processCount); err != nil {
		return err
	}

	s.envelopeStorageMetrics = NewEnvelopeStatCollector(s.control, s.Log)
	if err := s.registerer.Register(s.envelopeStorageMetrics); err != nil {
		return err
	}

	s.logStorageMetrics = NewLogStatCollector(s.control, s.Log)
	if err := s.registerer.Register(s.logStorageMetrics); err != nil {
		return err
	}

	sub, err := s.api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	s.updateAllCertificateExpiration()
	s.wg.Add(1)

	go s.run()
	return nil
}

func (s *Service) run() {
	defer s.wg.Done()
	for _, actorState := range s.actorMap.List() {
		s.updateActorOnlineStatus(actorState)
	}
	certificateExpirationCheckTicker := time.NewTicker(s.Opts.CertificateExpirationCheckInterval)
	defer certificateExpirationCheckTicker.Stop()

	certificateNotifyTicker := time.NewTicker(s.Opts.CertificateExpirationNotifyInterval)
	defer certificateNotifyTicker.Stop()

	for {
		select {
		case actorState := <-s.actorMap.Events:
			s.updateActorOnlineStatus(actorState)
		case <-certificateExpirationCheckTicker.C:
			s.updateAllCertificateExpiration()
		case <-certificateNotifyTicker.C:
			s.checkForExpirationNotify()
		case <-s.ctx.Done():
			return
		}
	}
}

// Shutdown the service
func (s *Service) Shutdown() error {
	s.registerer.Unregister(s.accountDaysToCertificateExpiration)
	s.registerer.Unregister(s.actorOnlineStatus)
	s.registerer.Unregister(s.serverDaysToCertificateExpiration)

	s.registerer.Unregister(s.processCount)
	s.registerer.Unregister(s.envelopeStorageMetrics)
	s.registerer.Unregister(s.logStorageMetrics)

	s.stop()
	s.wg.Wait()
	s.actorMap.Close()
	return nil
}

func (s *Service) updateActorOnlineStatus(actorState tools.ActorState) {
	if actorState.Kind != api.Actor_EMITTER {
		var value float64
		if actorState.Online {
			value = 1.0
		}
		s.actorOnlineStatus.WithLabelValues(actorState.Name).Set(value)
	}
}

func (s *Service) updateAllCertificateExpiration() {
	s.serverDaysToCertificateExpiration.
		WithLabelValues("server-ca").
		Set(time.Until(s.serverCAValidNotAfter).Hours() / 24)
	s.serverDaysToCertificateExpiration.
		WithLabelValues("client-ca").
		Set(time.Until(s.serverCAValidNotAfter).Hours() / 24)

	accountList, err := s.control.Account.List(control.AccountListRequest{})
	if err != nil {
		s.Opts.Logger.Err(err).Msg("Could not list accounts")
		return
	}
	for _, account := range accountList.AccountList {
		s.updateAccountDaysToCertificateExpiration(account)
	}
}

// ErrNoCert is returned by accountCertificateValidUntil if the account has no
// certificate
var ErrNoCert = errors.New("account has not certificate")

func accountCertificateValidUntil(account api.Account) (time.Time, error) {
	if account.Cert == "" {
		return time.Time{}, ErrNoCert
	}
	cert, err := xbus_tools.ReadPEMCertificate(account.Cert)
	if err != nil {
		return time.Time{}, fmt.Errorf(
			"Could not parse account %s certificate: %s", account.Name, err)
	}
	return cert.NotAfter, nil
}

func (s *Service) updateAccountDaysToCertificateExpiration(account api.Account) {
	validUntil, err := accountCertificateValidUntil(account)
	if err != nil {
		if err != ErrNoCert {
			s.Opts.Logger.Err(err).Msg("")
		}
		return
	}
	days := time.Until(validUntil).Hours() / 24
	s.accountDaysToCertificateExpiration.WithLabelValues(account.Name).Set(days)
}

func (s *Service) notifyCertificate(name string, notAfter time.Time) {
	remains := time.Until(notAfter)
	remainDays := int(math.Floor(remains.Hours() / 24))
	level := sentry.LevelInfo

	if remains <= 0 {
		level = sentry.LevelFatal
	} else if remains <= s.Opts.CertificateExpirationErrorThreshold {
		level = sentry.LevelError
	} else if remains <= s.Opts.CertificateExpirationWarningThreshold {
		level = sentry.LevelWarning
	}

	if level == sentry.LevelInfo {
		return
	}

	msg := fmt.Sprintf(
		"Certificate \"%s\" is about to expire. Time remaining: %d days", name, remainDays)

	if s.Opts.Hub != nil {
		event := sentry.NewEvent()
		event.Level = level
		event.Message = msg
		event.Extra = map[string]interface{}{
			"name":           name,
			"days-remaining": remainDays,
		}

		s.Opts.Hub.CaptureEvent(event)
	} else {
		if level == sentry.LevelWarning {
			s.Opts.Logger.Info().Msg(msg)
		} else {
			s.Opts.Logger.Error().Msg(msg)
		}
	}
}

func (s *Service) checkForExpirationNotify() {
	s.notifyCertificate("server-ca", s.serverCAValidNotAfter)
	s.notifyCertificate("client-ca", s.clientCAValidNotAfter)

	accountList, err := s.control.Account.List(control.AccountListRequest{})
	if err != nil {
		s.Opts.Logger.Err(err).Msg("Could not list accounts")
		return
	}
	for _, account := range accountList.AccountList {
		d, err := accountCertificateValidUntil(account)
		if err != nil {
			if err != ErrNoCert {
				s.Opts.Logger.Err(err).Msg("")
			}
			continue
		}
		s.notifyCertificate("Account "+account.Name, d)
	}
}

func (s *Service) handleCoreEventsProcess(evt api.CoreEvent) {
	s.TaskAdd()
	defer s.TaskDone()

	if evt.Process.Status != api.Process_DONE && evt.Process.Status != api.Process_ERROR {
		return
	}

	var (
		pipelineName    = evt.Process.PipelineID.String()
		pipelineVersion string
		actorName       = evt.Process.TriggerEmitterID.String()
	)
	if pipeline, err := s.pipelines.Get(evt.Process.PipelineID); err == nil {
		pipelineName = pipeline.Name
		pipelineVersion = pipeline.Version
	}
	if actor, err := s.actors.Get(evt.Process.TriggerEmitterID); err == nil {
		actorName = actor.Name
	}
	status := evt.Process.Status.String()

	// TODO we should get the actual process state changed instead of Now(),
	// which would be more accurate
	fmt.Println(time.Now())
	fmt.Println(evt.Process.CreatedAt)
	duration := time.Now().In(time.UTC).Sub(*evt.Process.CreatedAt)

	s.processCompletionTime.
		WithLabelValues(pipelineName, pipelineVersion, actorName, status).
		Observe(duration.Seconds())
}
