package metrics

import (
	"context"
	"fmt"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	"xbus.io/xbus/services/tools"
)

// NewSummaryCollector creates a new SummaryCollector
func NewSummaryCollector(
	control *control.Client, actors *tools.ActorCache, pipelines *tools.PipelineCache,
) *SummaryCollector {
	return &SummaryCollector{
		control:   control,
		actors:    actors,
		pipelines: pipelines,

		processCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "process", "count"),
			"Number of process in the bus",
			[]string{"pipeline", "pipeline_version", "emitter", "status", "result_acked"},
			nil,
		),

		storageCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "count"),
			"Number of items in the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "process"},
		),
	}
}

// SummaryCollector implements the prometheus.Collector interface
type SummaryCollector struct {
	control   *control.Client
	actors    *tools.ActorCache
	pipelines *tools.PipelineCache

	storageCountDesc *prometheus.Desc
	processCountDesc *prometheus.Desc
}

// Describe implements the prometheus.Collector interface.
func (c SummaryCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- c.processCountDesc
}

func (c *SummaryCollector) getActor(id api.UUID) string {
	a, err := c.actors.Get(id)
	if err != nil {
		return id.String()
	}
	return a.Name
}

func (c *SummaryCollector) getPipeline(id api.UUID) (string, string) {
	p, err := c.pipelines.Get(id)
	if err != nil {
		return id.String(), ""
	}
	return p.Name, p.Version
}

// Collect implements the prometheus.Collector interface.
func (c *SummaryCollector) Collect(ch chan<- prometheus.Metric) {
	ctx, cancel := context.WithTimeout(context.Background(), 4*time.Second)
	defer cancel()

	var count uint64

	if err := c.control.Process.QuerySummary(
		ctx, control.ProcessQueryRequest{},
		func(ctx context.Context, r control.ProcessSummary) {
			for _, entry := range r.Entries {
				count += entry.Count
				emitter := c.getActor(entry.EmitterID)
				pipeline, version := c.getPipeline(entry.PipelineID)
				ch <- prometheus.MustNewConstMetric(
					c.processCountDesc,
					prometheus.GaugeValue,
					float64(entry.Count),
					pipeline,
					version,
					emitter,
					entry.Status.String(),
					fmt.Sprintf("%v", entry.ResultAcked),
				)
			}
		},
	); err != nil {
		return
	}

	ch <- prometheus.MustNewConstMetric(
		c.storageCountDesc,
		prometheus.GaugeValue,
		float64(count),
		"default",
		"default",
	)
}
